{-----------------------------------------------------------------------------

 NW_Exec : All learn- and execute routines for the neural network.

 Consists of:

 Types:
 -TNetResult   : All possible net results

 Variables:
          Configuration variables :
 -WhenToCheck       : After how many direct good answers to check if all
                      examples are treated right
 -MaxDecreaseWeight : Max value to decrease weights with
 -MaxAwakenWeight   : Max value to increase weights with when dead
 -MaxIncreaseWeight : Max value to increase weights with
 -TamperIncrease    : How sensitive tampering is to increasing
 -InputStimulus     : Stimulus to give to input neuron when to activate
 -MaxWeight         : Maximum weight per link
 -MinWeight         : Minimum weight per link

 -MutateFactor         : After how many wrong tries times to mutate
 -MutateFlipFactor     : When mutating, chance to flip weight sign
 -MutateIncreaseFactor : When mutating, chance to increase weight
 -MutateDecreaseFactor : When mutating, chance to decrease weight

 -StepByStep : Operate in step-by-step mode or not

 -NetworkChanged : Whether or not the network has changed

 Routines:
 -LearnNetwork         : lets the network learn and such
 -NeuronStimulus       : Determines the stimulus the given Weight will be
                         going to give
 -NeuronActivated      : Determines whether the given Neuron was activated
 -RunNetwork           : runs the network and such
 -ClearNetwork         : Clears the network; set all values to non-used
                         and empty
 -GetRandomExample     : generates a random pick out of all examples
 -GiveNetworkImpuls    : Put a new impuls in the network's input neurons
 -GetNetworkOutput     : Gets the netork's output from it's output neurons
 -ActivateNetwork      : run the impuls set in the inputlayer through the net
 -WeakenNetwork        : Weakens all used links
 -TamperNetwork        : Tampers with all used links
 -StrengthenNetwork    : Strengthens all used links
 -DecreaseWeight       : give the new weight if the link has to weaken
 -AwakenedWeight       : give the new weight if the link has to awakened
 -IncreaseWeight       : give the new weight if the link has to strengthen
 -AwakenNetwork        : try to wake the network back to life if it is 'dead';
                         try to expand te signal tree in all branches
 -MutateNetwork        : dick around with the network (make some 'spontaneous'
                         mutations like in/decreases and sign flips)
 -GetNetResult         : Get the result by comparing the desired and actual
                         net output
 -SetDefExecPars       : sets the execute routine parameters to their defaults

-----------------------------------------------------------------------------}
unit NW_Exec;

interface

uses
  NW_List,
  NW_KIO,
  NW_Kernl,
  NW_Graph;

var
  { Configuration variables }
  WhenToCheck,               { After how many direct good answers to check if
                               all examples are treated right }
  MutateFactor,              { After how many wrong tries times to mutate }
  MutateFlipFactor,          { When mutating, chance to flip weight sign }
  MutateIncreaseFactor,      { When mutating, chance to increase weight }
  MutateDecreaseFactor,      { When mutating, chance to decrease weight }
  MaxDecreaseWeight,         { Max value to decrease weights with }
  MaxAwakenWeight,           { Max value to increase weights with when dead }
  MaxIncreaseWeight,         { Max value to increase weights with }
  TamperIncrease     : word; { How sensitive tampering is to increasing }
  InputStimulus,             { Stimulus to give to input neuron when to activate }
  MaxWeight,                     { Maximum weight per link }
  MinWeight          : integer;  { Minimum weight per link }
  StepByStep         : boolean;  { Operate in step-by-step mode or not }

var
  { And others }
  NetworkChanged     : boolean;  { Whether or not the network has changed }


{ LearnNetwork : lets the network learn and such }
procedure LearnNetwork;

{ Determines the stimulus the given Weight will be going to give }
function NeuronStimulus (Weight : integer) : integer;

{ Determines whether the given Neuron was activated }
function NeuronActivated (var Neuron : PNeuron) : boolean;

{ RunNetwork : runs the network and such }
procedure RunNetwork;

{ ClearNetwork : Clears the network; set all values to non-used and empty }
procedure ClearNetwork;

{ GetRandomExample : generates a random pick out of all examples }
function GetRandomExample : PExample;

{ GiveNetworkImpuls : Put a new impuls in the network's input neurons }
procedure GiveNetworkImpuls (InputValue : word);

{ GetNetworkOutput : Gets the netork's output from it's output neurons }
function GetNetworkOutput : word;

{ ActivateNetwork : run the impuls set in the inputlayer through the network }
procedure ActivateNetwork;

{ WeakenNetwork : Weakens all used links }
procedure WeakenNetwork;

{ TamperNetwork : Tampers with all used links }
procedure TamperNetwork;

{ StrengthenNetwork : Strengthens all used links }
procedure StrengthenNetwork;

{ DecreaseWeight : give the new weight if the link has to weaken }
function DecreasedWeight (OldWeight : integer) : integer;

{ AwakenedWeight : give the new weight if the link has to awakened }
function AwakenedWeight (OldWeight : integer) : integer;

{ IncreaseWeight : give the new weight if the link has to strengthen }
function IncreasedWeight (OldWeight : integer) : integer;

{ AwakenNetwork : try to wake the network back to life if it is 'dead';
                  try to expand te signal tree in all branches }
procedure AwakenNetwork;

{ MutateNetwork : dick around with the network (make some 'spontaneous'
                  mutations like in/decreases and sign flips) }
procedure MutateNetwork;

type
  { All possible net results }
  TNetResult = (Dead, Good, Reasonable, Bad);

const
  { ResultText : the text to display in the status bar while executing }
  ResultText : array [TNetResult] of string =
               ('Dead', 'Good', 'Reasonable', 'Bad');

{ GetNetResult : Get the result by comparing the desired and actual net output }
function GetNetResult (Output, Desired : word) : TNetResult;

{ SetDefExecPars : sets the execute routine parameters to their defaults }
procedure SetDefExecPars;

implementation

{ SetDefExecPars : sets the execute routine parameters to their defaults }
procedure SetDefExecPars;

begin
  WhenToCheck := 4;
  MutateFactor := 20;
  MutateFlipFactor := 10;
  MutateIncreaseFactor := 10;
  MutateDecreaseFactor := 10;
  MaxDecreaseWeight := 16;
  MaxAwakenWeight := 6;
  MaxIncreaseWeight := 30;
  TamperIncrease := 6;
  InputStimulus := 100;
  MaxWeight := 100;
  MinWeight := -50;
  StepByStep := false;
end; { SetDefExecPars }

{ GetRandomExample : generates a random pick out of all examples }
function GetRandomExample : PExample;

var
  InterExample : PExample;
  ExampleNr    : word;

begin
  ExampleNr := Random (AmExamples);
  InterExample := FirstExample;
  while ExampleNr <> 0 do begin
    InterExample := InterExample^.Next;
    ExampleNr := ExampleNr - 1;
  end; { while }
  GetRandomExample := InterExample;
end; { GetRandomExample }


{ ClearNetwork : Clears the network; set all values to non-used and empty }
procedure ClearNetwork;

var
  InterNeuron : PNeuron;
  InterLayer  : PLayer;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterNeuron^.Received := 0;
      InterNeuron^.Activated := false;
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        InterLink^.Used := false;
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { ClearNetwork }


{ GiveNetworkImpuls : Put a new impuls in the network's input neurons }
procedure GiveNetworkImpuls (InputValue : word);

var
  InterNeuron  : PNeuron;
  ProcessValue : word;

begin
  InterNeuron := FirstInputNeuron;
  ProcessValue := InputValue;
  while InterNeuron <> nil do begin
    InterNeuron^.Received := InputStimulus * (ProcessValue and 1);
    InterNeuron^.Activated := (InterNeuron^.Received > InterNeuron^.Treshold);
    InterNeuron := InterNeuron^.Next;
    ProcessValue := ProcessValue shr 1;
  end; { while }
end; { GiveNetworkImpuls }


{ GetNetworkOutput : Gets the network's output from it's output neurons }
function GetNetworkOutput : word;

var
  ProcessValue : word;
  InterNeuron  : PNeuron;

begin
  ProcessValue := 0;
  InterNeuron := LastOutputNeuron;
  while InterNeuron <> nil do begin
    if InterNeuron^.Activated then begin
      ProcessValue := ProcessValue or 1;
    end; { then }
    ProcessValue := ProcessValue shl 1;
    InterNeuron := InterNeuron^.Prev;
  end; { while }
  ProcessValue := ProcessValue shr 1;

  GetNetworkOutput := ProcessValue;
end; { GetNetworkOutput }

{ Determines the delta of the stimulus the given Weight will be giving }
function DNeuronStimulus (Weight : integer) : integer;

begin
  DNeuronStimulus := 1;
{
 -----Other possibilities:-----
  DNeuronStimulus :=
   Round (15 / (exp (15 * (Weight / MaxWeight - 0.5)) + 2 + exp (-15 * (Weight / MaxWeight - 0.5))) +
          15 / (exp (15 * (Weight / MinWeight - 0.5)) + 2 + exp (-15 * (Weight / MinWeight - 0.5))));
}
end; { DNeuronStimulus }

{ Determines the stimulus the given Weight will be going to give }
function NeuronStimulus (Weight : integer) : integer;

begin
  NeuronStimulus := Weight;
{
 -----Other possibilities:-----
  NeuronStimulus :=
   Round ((2 / (1 + exp (-10 * Weight / MaxWeight)) - 1) * MaxWeight);
  NeuronStimulus :=
   Round (MaxWeight / (1 + exp (-15 * (Weight / MaxWeight - 0.5))) +
          MinWeight / (1 + exp (-15 * (Weight / MinWeight - 0.5))));
}
end; { NeuronStimulus }

{ Determines whether the given Neuron was activated }
function NeuronActivated (var Neuron : PNeuron) : boolean;

begin
  NeuronActivated := (Neuron^.Received >= Neuron^.Treshold);
end; { NeuronActivated }

{ ActivateNetwork : run the impuls set in the inputlayer through the network }
procedure ActivateNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  { Start at the first layer }
  InterLayer := FirstLayer;

  { And for every layer }
  while InterLayer^.Next <> nil do begin
    { Start at the first neuron }
    InterNeuron := InterLayer^.FirstNeuron;

    { And for every neuron in the layer do }
    while InterNeuron <> nil do begin
      { Look if it was activated }
      InterNeuron^.Activated := NeuronActivated (InterNeuron);

      { If the neuron wants to transmit }
      if InterNeuron^.Activated then begin
        { Then start at the first link to the next layer }
        InterLink := InterNeuron^.FirstLink;

        { And for every link }
        while InterLink <> nil do begin
          { Give a stimulus to the receiving neuron }
          InterLink^.Neuron^.Received := InterLink^.Neuron^.Received +
                                         NeuronStimulus (InterLink^.Weight);
          InterLink^.Used := true;

          { And take the next link }
          InterLink := InterLink^.Next;
        end; { while }
      end; { then }

      { And take the next neuron }
      InterNeuron := InterNeuron^.Next;
    end; { while }

    { And go to the next layer }
    InterLayer := InterLayer^.Next;
  end; { while }

  { And look which outputneurons were activated }
  InterNeuron := FirstOutputNeuron;
  while InterNeuron <> nil do begin
    InterNeuron^.Activated := NeuronActivated (InterNeuron);
    InterNeuron := InterNeuron^.Next;
  end; { while }
end; { ActivateNetwork }

{ IncreaseWeight : give the new weight if the link has to strengthen }
function IncreasedWeight (OldWeight : integer) : integer;

var
  NewWeight : integer;

begin
  if OldWeight < 0 then begin
    NewWeight := OldWeight
                 - DNeuronStimulus (OldWeight)
                 - Random (MaxIncreaseWeight);
    if NewWeight < MinWeight then begin
      NewWeight := MinWeight;
    end; { then }
  end { then }
  else begin
    NewWeight := OldWeight
                 + DNeuronStimulus (OldWeight)
                 + Random (MaxIncreaseWeight);
    if NewWeight > MaxWeight then begin
      NewWeight := MaxWeight;
    end; { then }
  end; { else }
  IncreasedWeight := NewWeight;
end; { IncreasedWeight }


{ AwakenedWeight : give the new weight if the link has to awakened }
function AwakenedWeight (OldWeight : integer) : integer;

var
  NewWeight : integer;

begin
  NewWeight := OldWeight
               {+ Abs (DNeuronStimulus (OldWeight))}
               + Random (MaxAwakenWeight);
  if NewWeight > MaxWeight then begin
    NewWeight := MaxWeight;
  end; { then }
  AwakenedWeight := NewWeight;
end; { AwakenedWeight }


{ DecreaseWeight : give the new weight if the link has to weaken }
function DecreasedWeight (OldWeight : integer) : integer;

var
  NewWeight : integer;

begin
  if OldWeight > 0 then begin
    NewWeight := OldWeight
                 - DNeuronStimulus (OldWeight)
                 - Random (MaxDecreaseWeight);
    if NewWeight < MinWeight then begin
      NewWeight := MinWeight;
    end; { then }
  end { then }
  else begin
    NewWeight := OldWeight
                 + DNeuronStimulus (OldWeight)
                 + Random (MaxDecreaseWeight);
    if NewWeight > MaxWeight then begin
      NewWeight := MaxWeight;
    end; { then }
  end; { else }
  DecreasedWeight := NewWeight;
end; { DecreasedWeight }


{ StrengthenNetwork : Strengthens the net. All used linkes with positive
                      weights and result, as well as negative weights without
                      results are strengthened, all other links are weakened }
procedure StrengthenNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer^.Next <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        if InterLink^.Used then begin
          if ((InterLink^.Weight >= 0) and InterLink^.Neuron^.Activated) or
             ((InterLink^.Weight < 0) and not InterLink^.Neuron^.Activated) then begin
            InterLink^.Weight := IncreasedWeight (InterLink^.Weight);
          end; { then }
        end; { then }
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { StrengthenNetwork }

{ TamperNetwork : Tampers with the net. All used links with effect are either
                  strengthened or weakened }
procedure TamperNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer^.Next <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        if InterLink^.Used then begin
          if ((InterLink^.Weight >= 0) and InterLink^.Neuron^.Activated) or
             ((InterLink^.Weight < 0) and not InterLink^.Neuron^.Activated) then begin
            if Random (TamperIncrease) = 0 then begin
              InterLink^.Weight := IncreasedWeight (InterLink^.Weight);
            end { then }
            else begin
              InterLink^.Weight := DecreasedWeight (InterLink^.Weight);
            end; { else }
          end; { then }
        end; { then }
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { TamperNetwork }


{ WeakenNetwork : Weakens the net. All used links with effect are weakened }
procedure WeakenNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer^.Next <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        if InterLink^.Used then begin
          if ((InterLink^.Weight >= 0) and InterLink^.Neuron^.Activated) or
             ((InterLink^.Weight < 0) and not InterLink^.Neuron^.Activated) then begin
            InterLink^.Weight := DecreasedWeight (InterLink^.Weight);
          end; { then }
        end; { then }
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { WeakenNetwork }

{ AwakenNetwork : try to wake the network back to life if it is 'dead';
                  try to expand the signal tree in all branches (only +);
                  increase all used links which had no effect }
procedure AwakenNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer^.Next <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        if InterLink^.Used then begin
          if not ((InterLink^.Weight >= 0) and InterLink^.Neuron^.Activated) and
             not ((InterLink^.Weight < 0) and not InterLink^.Neuron^.Activated) then begin
            InterLink^.Weight := AwakenedWeight (InterLink^.Weight);
          end; { then }
        end; { then }
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { AwakenNetwork }


{ MutateNetwork : dick around with the network (make some 'spontaneous'
                  mutations like in/decreases and sign flips) }
procedure MutateNetwork;

var
  InterLayer  : PLayer;
  InterNeuron : PNeuron;
  InterLink   : PNeuronLinker;

begin
  InterLayer := FirstLayer;
  while InterLayer^.Next <> nil do begin
    InterNeuron := InterLayer^.FirstNeuron;
    while InterNeuron <> nil do begin
      InterLink := InterNeuron^.FirstLink;
      while InterLink <> nil do begin
        if Random (MutateIncreaseFactor) = 1 then begin
          InterLink^.Weight := IncreasedWeight (InterLink^.Weight);
        end; { else then }
        if Random (MutateDecreaseFactor) = 1 then begin
          InterLink^.Weight := DecreasedWeight (InterLink^.Weight);
        end; { else then }
        if Random (MutateFlipFactor) = 1 then begin
          InterLink^.Weight := -InterLink^.Weight;
          if InterLink^.Weight < MinWeight then begin
            InterLink^.Weight := MinWeight;
          end; { then }
          if InterLink^.Weight > MaxWeight then begin
            InterLink^.Weight := MaxWeight;
          end; { then }
        end; { else then }
        InterLink := InterLink^.Next;
      end; { while }
      InterNeuron := InterNeuron^.Next;
    end; { while }
    InterLayer := InterLayer^.Next;
  end; { while }
end; { MutateNetwork }


{ GetNetResult : Get the result by comparing the desired and actual net output }
function GetNetResult (Output, Desired : word) : TNetResult;

begin
  { Look if 'dead' result is desirable }
  if Desired = 0 then begin
    { Yes -> look if 'dead' result given by net }
    if Output = Desired then begin
      GetNetResult := Good;
    end { then }
    else begin
      GetNetResult := Bad;
    end; { else }
  end { then }
  else begin
    { No -> look what net did }
    if Output = Desired then begin
      { Good output }
      GetNetResult := Good;
    end { then }
    else if Output = 0 then begin
      { Dead network }
      GetNetResult := Dead;
    end { else if }
    else if (Output and Desired) = Desired then begin
      { Some matches made }
      GetNetResult := Reasonable;
    end { else if }
    else begin
      { Totally wrong output given }
      GetNetResult := Bad;
    end; { else }
  end; { else }
end; { GetNetResult }


{ ExampleResult : runs the given example through the network,
                  and gives it's result }
function ExampleResult (Example : PExample) : TNetResult;

var
  NetWorkOutput : word;

begin
  { Clear the network }
  ClearNetwork;

  { Set the initial value in the input neurons }
  GiveNetworkImpuls (Example^.Input);

  { And run the impuls through the network }
  ActivateNetwork;

  { Get the output value of the net }
  NetworkOutput := GetNetworkOutput;

  { Compare the given output with the desired output }
  ExampleResult := GetNetResult (NetworkOutput, Example^.Output);
end; { ExampleResult }




{ CheckAllExamples : checks how many examples are treated the right way }
function CheckAllExamples : word;

var
  Result       : TNetResult;
  AmGood       : word;
  InterExample : PExample;

begin
  InterExample := FirstExample;
  AmGood := 0;
  while InterExample <> nil do begin
    Result := ExampleResult (InterExample);
    if Result = Good then begin
      AmGood := AmGood + 1;
    end; { then }
    InterExample := InterExample^.Next;
  end; { ehile }

  CheckAllExamples := AmGood;
end; { CheckAllExamples }






{ RunNetwork : runs the network and such }
procedure RunNetwork;

var
  Example       : PExample;
  NetworkOutput : word;
  Input         : string;
  UserChoice    : char;

begin
  { Look if a network is in memory }
  if not NetworkInMemory then begin
    { No -> then how to run?!?!? }
    ClearScreen;
    CursorPos := Middle;
    ShowText ('There has to be at least one neuron in memory before it can be run . . .');
    ScrollDown;
    ShowText ('Press any key to return to main menu . . .');
    WaitKeyPress;
    Exit;
  end; { then }

  repeat
    { Get the to-use example from the user }
    ClearScreen;
    CursorPos := Home;
    ShowText ('Running network with user-fed examples.');
    ScrollDown;
    ShowText ('Give the desired input value as a binary sequence:');

    Input := GetText ('');

    Example^.Input := MakeTextExample (Input);

    { Clear the network }
    ClearNetwork;

    { Set the initial value in the input neurons }
    GiveNetworkImpuls (Example^.Input);

    { And run the impuls through the network }
    ActivateNetwork;

    { Get the output value of the net }
    NetworkOutput := GetNetworkOutput;

    { And show the user the net's output }
    GoBeginLine;
    ScrollDown;
    ScrollDown;
    ShowText ('Inserted in net : ' + MakeExampleText (Example^.Input, AmInput, AmOutput, IsInput));
    ShowText ('Given in return : ' + MakeExampleText (NetworkOutput, AmInput, AmOutput, IsOutput));
    ScrollDown;
    ShowText ('Press a key to feed in next try (or Esc to return to main menu) . . .');
    DeleteCursor;

    UserChoice := GetKey;
  until UserChoice = #27;
end; { RunNetwork }




{ LearnNetwork : lets the network learn and such }
procedure LearnNetwork;

var
  { The network's output + status }
  AmGood,
  AmContinuousGood,
  AmReasonable,
  AmBad,
  AmDead,
  AmWrongSinceLastGood,
  Step          : LongInt;
  Result        : TNetResult;

  { Routine display + stop control }
  InChar   : char;
  TextLine : string;
  Finished : boolean;

  { The desired example for learning + learning control }
  Example        : PExample;
  AllExamplesLearned,
  ExampleLearned : boolean;

begin
  { Look if a network is in memory }
  if not NetworkInMemory then begin
    { No -> then how to run?!?!? }
    ClearScreen;
    CursorPos := Middle;
    NewColor (15);
    ShowText ('There has to be at least one neuron in memory before it can be run . . .');
    ScrollDown;
    ShowText ('Press any key to return to main menu . . .');
    WaitKeyPress;
    Exit;
  end; { then }

  { Show instructions }
  ClearScreen;
  CursorPos := Home;
  NewColor (15);
  ShowText ('When simulating, press:');
  ShowText ('-SPACE  to go to another display mode;');
  ShowText ('-ESCAPE to quit the run;');
  ShowText ('-ENTER  to switch between step-by-step mode and full free run.');
  ScrollDown;
  ShowText ('When in step-by-step mode, press any other key than the one mentioned above to');
  ShowText ('go to the next step.');
  ScrollDown;
  ShowText ('Press any key to start the simulation.');
  DeleteCursor;
  WaitKeyPress;

  { And start the learning process }
  AllExamplesLearned := false;
  ExampleLearned := true;
  Finished := false;
  ClearScreen;
  ShowOptions (StepByStep);
  CursorPos := Home;
  AmGood := 0;
  AmWrongSinceLastGood := 0;
  AmContinuousGood := 0;
  AmReasonable := 0;
  AmBad := 0;
  AmDead := 0;
  Step := 0;
  repeat
    { Look if new input/output is desired }
    if ExampleLearned then begin
      { Get a new example to learn }
      Example := GetRandomExample;
      ExampleLearned := false;
      AmWrongSinceLastGood := 0;
    end; { then }

    { Run the network with the selected example }
    Result := ExampleResult (Example);

    { And take the right measurements }
    case Result of
      Dead       : begin
                     AwakenNetwork;
                     AmDead := AmDead + 1;
                   end;
      Good       : begin
                     StrengthenNetwork;
                     ExampleLearned := true;
                     AmGood := AmGood + 1;
                   end; { case Good }
      Reasonable : begin
                     TamperNetwork;
                     AmReasonable := AmReasonable + 1;
                   end;
      Bad        : begin
                     WeakenNetwork;
                     AmBad := AmBad + 1;
                   end;
    end; { case }

    { Look if net did it all right in one try }
    if ExampleLearned then begin
      if AmWrongSinceLastGood = 0 then begin
        AmContinuousGood := AmContinuousGood + 1;
        if AmContinuousGood > WhenToCheck then begin
          if DisplayMode = DisplayText then begin
            NewColor (15);
            ShowText ('Checking all possible examples . . .                                          ');
          end; { then }
          AllExamplesLearned := (CheckAllExamples = AmExamples);
        end; { then }
      end { then }
      else begin
        AmContinuousGood := 0;
      end; { else }
    end; { then }

    { And look if a mutation needs to happen }
    if not ExampleLearned then begin
      AmWrongSinceLastGood := AmWrongSinceLastGood + 1;
    end; { then }
    if AmWrongSinceLastGood > MutateFactor then begin
      MutateNetwork;
      if DisplayMode = DisplayText then begin
        NewColor (15);
        ShowText ('Mutating network . . .                                                        ');
      end; { then }
      AmWrongSinceLastGood := 0;
    end; { then }
    Step := Step + 1;

    { Update the user on the current situation }
    if DisplayMode = DisplayGraph then begin
      DrawNeurons;
      GiveUsedLinks;
      GiveValues;
      ShowStatus (ResultText [Result], Step, AmGood, AmReasonable, AmBad, AmDead);
    end { then }
    else if DisplayMode = DisplayText then begin
      Str (Step, TextLine);
      TextLine := 'Run: ' + TextLine +
                  '; in: ' + MakeExampleText (Example^.Input, AmInput, AmOutput, IsInput) +
                  '; out: ' + MakeExampleText (GetNetworkOutput, AmInput, AmOutput, IsOutput) +
                  '; wanted: ' + MakeExampleText (Example^.Output, AmInput, AmOutput, IsOutput) + ' > ';
      case Result of
        Dead       : begin
                       TextLine := TextLine + 'Dead; awakening net';
                       NewColor (7);
                     end;
        Good       : begin
                       TextLine := TextLine + 'Good; strengthening net';
                       NewColor (10);
                     end;
        Reasonable : begin
                       TextLine := TextLine + 'Undef.; tampering with net';
                       NewColor (14);
                     end;
        Bad        : begin
                       TextLine := TextLine + 'Bad; weakening net';
                       NewColor (12);
                     end;
      end; { case }
      ShowText (TextLine + '          ');
      ShowStatus (ResultText [Result], Step, AmGood, AmReasonable, AmBad, AmDead);
    end; { else then }

    { Look if inter-step delay is required }
    if StepByStep and not AllExamplesLearned then begin
      WaitKeyPressNoRead;
    end; { then }

    { And look if user gets bored }
    if not KBQueueEmpty then begin
      InChar := GetKey;
      case InChar of
        ' ' : begin
                { Take next display mode }
                NextDisplayMode;

                { Show the current options }
                ClearScreen;
                ShowOptions (StepByStep);

                { And do some specialties }
                if DisplayMode = DisplayText then begin
                  CursorPos := Home;
                end; { then }
              end;
        #27 : begin
                Finished := true;
              end;
        #13 : begin
                { Go to the other stepmode }
                StepByStep := not StepByStep;

                { And show the current options }
                ShowOptions (StepByStep);
              end;
      end; { case }
    end; { then }
  until Finished or AllExamplesLearned;
  NetworkChanged := true;

  { Update user on exiting cause }
  ClearScreen;
  CursorPos := Middle;
  NewColor (15);
  if Finished then begin
    { Exiting due to user intervention }
    ShowText ('Network learning process halted on user''s request.');
    ScrollDown;
  end { then }
  else begin
    ShowText ('Network fully trained; all examples treated correctly.');
    ScrollDown;
  end; { else }

  ShowText ('Press a key to return to the main menu.');
  DeleteCursor;
  WaitKeyPress;
end; { LearnNetwork }

begin
end. { NW_Exec }